// ode3dsloader-0.0 by Kosei Demura 2007-11-01
// main.cpp 
// 3dstH[}bgōꂽIuWFNgǂݍރvO
// lib3dsCuKv@http://lib3ds.sourceforge.net/
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "ode3dsloader.h"

extern int MeshCount;

// #define DEBUG

/* for VC++, no precision loss complaints */
#if _MSC_VER
#pragma warning(disable:4244 4305) 
#endif

#ifdef dDOUBLE
#define dsDrawBox      dsDrawBoxD
#define dsDrawSphere   dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawLine     dsDrawLineD
#define dsDrawTriangle dsDrawTriangleD
#endif

#define MAX_CONTACTS 50

typedef struct MyObject {
  dBodyID body;	       
  dGeomID geom;	
  dReal   m; 
  dReal   lx,ly,lz; 

  // Trimesh only - double buffered matrices for 'last transform' setup
  dReal matrix_dblbuff[ 16 * 2 ];
  int last_matrix_index;  
} MyObject;

static dsFunctions fn;
static dWorldID world;
static dSpaceID space;
static MyObject *model = 0;
static dGeomID  ground;
static dJointGroupID contactgroup;
static dTriMeshDataID *TriData = 0; 

typedef struct 
{
    float *vertices;
    unsigned int vertexCount;
    int *indices;
    unsigned int indexCount;
} MyTrimesh;

MyTrimesh *trimesh = 0;

// set previous transformation matrix for trimesh
void setCurrentTransform(dGeomID geom)
{
 const dReal *Pos = dGeomGetPosition(geom);
 const dReal *Rot = dGeomGetRotation(geom);

 const dReal Transform[16] = 
 {
   Rot[0], Rot[4], Rot[8],  0,
   Rot[1], Rot[5], Rot[9],  0,
   Rot[2], Rot[6], Rot[10], 0,
   Pos[0], Pos[1], Pos[2],  1
 };

 dGeomTriMeshSetLastTransform(geom, *(dMatrix4*)(&Transform) );
}


static void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
  dBodyID b1 = dGeomGetBody(o1);
  dBodyID b2 = dGeomGetBody(o2);
  if (b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeContact)) return;

  dContact contact[MAX_CONTACTS];  
  for (int i=0; i<MAX_CONTACTS; i++) {
    contact[i].surface.mode = dContactBounce | dContactSoftCFM | dContactSoftERP;
    contact[i].surface.mu         = dInfinity; // 1.0; 
    contact[i].surface.bounce     = 0.1;
    contact[i].surface.soft_cfm   = 0.01;
    contact[i].surface.soft_erp   = 0.2;
  }
  if (int numc = dCollide (o1,o2,MAX_CONTACTS,&contact[0].geom,
			   sizeof(dContact))) {
    dMatrix3 RI;
    dRSetIdentity(RI);
    for (int i=0; i<numc; i++) {
      dJointID c = dJointCreateContact(world,contactgroup,contact+i);
      dJointAttach(c,b1,b2);
    }
  }
}

void drawModel()
{
  dsSetColor(1.1, 1.1, 1.1);

	for (int j = 0; j < MeshCount; j++) {

		// assume all trimeshes are drawn as bunnies
    const dReal *Pos = dGeomGetPosition(model[j].geom);
    const dReal *Rot = dGeomGetRotation(model[j].geom);
					
	  //for (int i = 0; i < trimesh[j].indexCount /3 ; i++){
  	for (int i = 0; i < (int) trimesh[j].indexCount; i++){
	  	const dReal v[9] = {
      	trimesh[j].vertices[trimesh[j].indices[0+i*3] * 3 + 0],
      	trimesh[j].vertices[trimesh[j].indices[0+i*3] * 3 + 1],
      	trimesh[j].vertices[trimesh[j].indices[0+i*3] * 3 + 2],
      	trimesh[j].vertices[trimesh[j].indices[1+i*3] * 3 + 0],
      	trimesh[j].vertices[trimesh[j].indices[1+i*3] * 3 + 1],
      	trimesh[j].vertices[trimesh[j].indices[1+i*3] * 3 + 2],
      	trimesh[j].vertices[trimesh[j].indices[2+i*3] * 3 + 0],
      	trimesh[j].vertices[trimesh[j].indices[2+i*3] * 3 + 1],
      	trimesh[j].vertices[trimesh[j].indices[2+i*3] * 3 + 2]
	  	};
    	dsDrawTriangle(Pos, Rot,&v[0], &v[3], &v[6], 1);
  	}

 		// tell the tri-tri collider the current transform of the trimesh --
   	// this is fairly important for good results.          
	  // Fill in the (4x4) matrix.
	  dReal *p_matrix = model[j].matrix_dblbuff + (model[j].last_matrix_index * 16);

	  p_matrix[ 0] = Rot[ 0];	p_matrix[ 1] = Rot[ 1];	p_matrix[ 2] = Rot[ 2];	p_matrix[ 3] = 0;
	  p_matrix[ 4] = Rot[ 4];	p_matrix[ 5] = Rot[ 5];	p_matrix[ 6] = Rot[ 6];	p_matrix[ 7] = 0;
	  p_matrix[ 8] = Rot[ 8];	p_matrix[ 9] = Rot[ 9];	p_matrix[10] = Rot[10];	p_matrix[11] = 0;
	  p_matrix[12] = Pos[ 0];	p_matrix[13] = Pos[ 1];	p_matrix[14] = Pos[ 2];	p_matrix[15] = 1;

	  // Flip to other matrix.
	  model[j].last_matrix_index = !model[j].last_matrix_index;

	  dGeomTriMeshSetLastTransform( model[j].geom, 
		  *(dMatrix4*)( model[j].matrix_dblbuff + model[j].last_matrix_index * 16 ));
  }
}

static void simLoop(int pause)
{
  if (!pause) {
		dSpaceCollide(space,0,&nearCallback);

  	for (int i=0; i< MeshCount; i++) {
    	if (model[i].geom) {
      	if (dGeomGetClass(model[i].geom) == dTriMeshClass) 
						setCurrentTransform(model[i].geom); 
   		}
  	}

		dWorldStep(world,0.01);	
		dJointGroupEmpty(contactgroup); 
	}

  drawModel(); 
}

void makeModel() 
{
  static dReal weight = 10.0;
  dReal x0 = 0, y0 = 0, z0 = 1.0;
	dMass m;


	load3dsModel("../models/rabbit.3DS");

	TriData = (dTriMeshDataID *) malloc(MeshCount * sizeof(dTriMeshDataID));

	model  = (MyObject *) malloc (MeshCount * sizeof(MyObject));

	for (int i = 0; i < MeshCount; i++) {

		// rigid body
		model[i].body = dBodyCreate(world);
		dBodySetPosition(model[i].body, x0, y0, z0);

		// for a trimesh object
  	TriData[i] = dGeomTriMeshDataCreate();	  

    dGeomTriMeshDataBuildSingle(TriData[i], trimesh[i].vertices, 3 * sizeof(float), 
	     trimesh[i].vertexCount, trimesh[i].indices, trimesh[i].indexCount * 3, 3 * sizeof(int));
  
  	model[i].geom = dCreateTriMesh(space, TriData[i], 0, 0, 0);

		// remember the mesh's dTriMeshDataID on its userdata for convenience.
		dGeomSetData(model[i].geom, TriData[i]);  	 

		dMassSetTrimeshTotal(&m, weight, model[i].geom);
		// printf("mass at %f %f %f\n", m.c[0], m.c[1], m.c[2]);

  	dGeomSetPosition(model[i].geom, -m.c[0], -m.c[1], -m.c[2]);
    dMassTranslate(&m, -m.c[0], -m.c[1], -m.c[2]);

		dGeomSetBody(model[i].geom, model[i].body);
		dBodySetMass(model[i].body, &m);
 	
		dMatrix3 Rotation;
    dRFromAxisAndAngle(Rotation, 0, 0, 1, -M_PI/4);
  	dBodySetRotation(model[i].body, Rotation);
	} 
}

void command(int cmd)
{
	float xyz[3],hpr[3];

	switch(cmd){
		case' ':
			dsGetViewpoint( xyz, hpr );
			printf( "xyz=%4.2f %4.2f %4.2f ", xyz[0], xyz[1], xyz[2] );
			printf( "hpr=%6.2f %6.2f %5.2f \n", hpr[0], hpr[1], hpr[2] );
			break;
	}
}

static void start()
{
  static float xyz[3] = { 1.48, 1.14, 0.43};
  static float hpr[3] = { -140, -5.5, 0.00};
  printf("Space - Get view point\n");
  dsSetViewpoint (xyz,hpr);
}


void setDrawStuff()
{
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = &command;
  fn.path_to_textures = "./textures";
}

int main(int argc, char **argv) 
{
  setDrawStuff();
  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0);
  ground = dCreatePlane(space,0,0,1,0);
  dWorldSetGravity(world,0,0,-9.8);

  makeModel();
  dsSimulationLoop(argc,argv,640,400,&fn);

  dJointGroupDestroy(contactgroup);
  dSpaceDestroy(space);
  dWorldDestroy(world);
  return 0;
}
